@@ -35,14 +35,28 @@ static xmlXPathCompExprPtr models_xpath;
 static xmlXPathCompExprPtr layouts_xpath;
 static xmlXPathCompExprPtr option_groups_xpath;
 
+static GRegex **xml_encode_regexen = NULL;
+static GRegex **xml_decode_regexen = NULL;
+static const char *xml_decode_regexen_str[] = { "&lt;", "&gt;", "&amp;" };
+static const char *xml_encode_regexen_str[] = { "<", ">", "&" };
+
 /* gettext domain for translations */
 #define XKB_DOMAIN "xkeyboard-config"
 
 enum {
 	PROP_0,
-	PROP_ENGINE,
+	PROP_ENGINE
 };
 
+typedef struct {
+	gchar **patterns;
+	TwoConfigItemsProcessFunc func;
+	gpointer data;
+	gboolean country_matched;
+	gboolean language_matched;
+	const XklConfigItem *layout_item;
+} SearchParamType;
+
 static gboolean
 xkl_xml_find_config_item_child(xmlNodePtr iptr, xmlNodePtr * ptr)
 {
@@ -71,7 +85,7 @@ xkl_xml_find_config_item_child(xmlNodePtr iptr, xmlNodePtr * ptr)
 }
 
 static xmlNodePtr
-xkl_find_nonlocalized_element(xmlNodePtr ptr, const gchar * tag_name)
+xkl_find_element(xmlNodePtr ptr, const gchar * tag_name)
 {
 	xmlNodePtr found_element = NULL;
 
@@ -79,14 +93,9 @@ xkl_find_nonlocalized_element(xmlNodePtr ptr, const gchar * tag_name)
 	while (ptr != NULL) {
 		char *node_name = (char *) ptr->name;
 		if (ptr->type != XML_TEXT_NODE) {
-			xmlChar *lang = xmlNodeGetLang(ptr);
-
-			if (lang == NULL) {	/* No language specified */
-				if (!g_ascii_strcasecmp
-				    (node_name, tag_name)) {
-					found_element = ptr;
-					break;
-				}
+			if (!g_ascii_strcasecmp(node_name, tag_name)) {
+				found_element = ptr;
+				break;
 			}
 		}
 		ptr = ptr->next;
@@ -101,7 +110,7 @@ xkl_item_populate_optional_array(XklConfigItem * item, xmlNodePtr ptr,
 				 const gchar property_name[])
 {
 	xmlNodePtr top_list_element =
-	    xkl_find_nonlocalized_element(ptr, list_tag), element_ptr;
+	    xkl_find_element(ptr, list_tag), element_ptr;
 	gint n_elements, idx;
 	gchar **elements = NULL;
 
@@ -114,7 +123,7 @@ xkl_item_populate_optional_array(XklConfigItem * item, xmlNodePtr ptr,
 	element_ptr = top_list_element->children;
 	while (NULL !=
 	       (element_ptr =
-		xkl_find_nonlocalized_element(element_ptr, element_tag))) {
+		xkl_find_element(element_ptr, element_tag))) {
 		n_elements++;
 		element_ptr = element_ptr->next;
 	}
@@ -127,8 +136,7 @@ xkl_item_populate_optional_array(XklConfigItem * item, xmlNodePtr ptr,
 	element_ptr = top_list_element->children;
 	for (idx = 0;
 	     NULL != (element_ptr =
-		      xkl_find_nonlocalized_element
-		      (element_ptr, element_tag));
+		      xkl_find_element(element_ptr, element_tag));
 	     element_ptr = element_ptr->next, idx++) {
 		elements[idx] =
 		    g_strdup((const char *) element_ptr->
@@ -141,15 +149,20 @@ xkl_item_populate_optional_array(XklConfigItem * item, xmlNodePtr ptr,
 	return TRUE;
 }
 
+#include "libxml/parserInternals.h"
+
 gboolean
-xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
-		     XklConfigItem * item)
+xkl_read_config_item(XklConfigRegistry * config, gint doc_index,
+		     xmlNodePtr iptr, XklConfigItem * item)
 {
 	xmlNodePtr name_element, ptr;
 	xmlNodePtr desc_element = NULL, short_desc_element =
 	    NULL, vendor_element = NULL;
 
-	gchar *vendor = NULL;
+	gchar *vendor = NULL, *translated = NULL, *escaped =
+	    NULL, *unescaped = NULL;
+
+	gint i;
 
 	*item->name = 0;
 	*item->short_description = 0;
@@ -162,6 +175,10 @@ xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
 	if (!xkl_xml_find_config_item_child(iptr, &ptr))
 		return FALSE;
 
+	if (doc_index > 0)
+		g_object_set_data(G_OBJECT(item), XCI_PROP_EXTRA_ITEM,
+				  GINT_TO_POINTER(TRUE));
+
 	ptr = ptr->children;
 
 	if (ptr->type == XML_TEXT_NODE)
@@ -169,11 +186,9 @@ xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
 	name_element = ptr;
 	ptr = ptr->next;
 
-	short_desc_element =
-	    xkl_find_nonlocalized_element(ptr, XML_TAG_SHORT_DESCR);
-	desc_element = xkl_find_nonlocalized_element(ptr, XML_TAG_DESCR);
-	vendor_element =
-	    xkl_find_nonlocalized_element(ptr, XML_TAG_VENDOR);
+	short_desc_element = xkl_find_element(ptr, XML_TAG_SHORT_DESCR);
+	desc_element = xkl_find_element(ptr, XML_TAG_DESCR);
+	vendor_element = xkl_find_element(ptr, XML_TAG_VENDOR);
 
 	if (name_element != NULL && name_element->children != NULL)
 		strncat(item->name,
@@ -183,17 +198,49 @@ xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
 	if (short_desc_element != NULL
 	    && short_desc_element->children != NULL) {
 		strncat(item->short_description,
-			dgettext(XKB_DOMAIN,
-				 (const char *) short_desc_element->
-				 children->content),
+			dgettext(XKB_DOMAIN, (const char *)
+				 short_desc_element->children->content),
 			XKL_MAX_CI_SHORT_DESC_LENGTH - 1);
 	}
 
 	if (desc_element != NULL && desc_element->children != NULL) {
+		/* Convert all xml-related characters to XML form, otherwise dgettext won't find the translation 
+		 * The conversion is not using libxml2, because there are no handy functions in API */
+		translated =
+		    g_strdup((gchar *) desc_element->children->content);
+		for (i =
+		     sizeof(xml_encode_regexen_str) /
+		     sizeof(xml_encode_regexen_str[0]); --i >= 0;) {
+			escaped =
+			    g_regex_replace(xml_encode_regexen[i],
+					    translated, -1, 0,
+					    xml_decode_regexen_str[i], 0,
+					    NULL);
+			g_free(translated);
+			translated = escaped;
+		}
+		escaped = translated;
+
+		/* Do the translation! */
+		translated =
+		    g_strdup(dgettext(XKB_DOMAIN, (const char *) escaped));
+		g_free(escaped);
+
+		/* Convert all XML entities back to normal form */
+		for (i =
+		     sizeof(xml_decode_regexen_str) /
+		     sizeof(xml_decode_regexen_str[0]); --i >= 0;) {
+			unescaped =
+			    g_regex_replace(xml_decode_regexen[i],
+					    translated, -1, 0,
+					    xml_encode_regexen_str[i], 0,
+					    NULL);
+			g_free(translated);
+			translated = unescaped;
+		}
 		strncat(item->description,
-			dgettext(XKB_DOMAIN,
-				 (const char *) desc_element->children->
-				 content), XKL_MAX_CI_DESC_LENGTH - 1);
+			translated, XKL_MAX_CI_DESC_LENGTH - 1);
+		g_free(translated);
 	}
 
 	if (vendor_element != NULL && vendor_element->children != NULL) {
@@ -216,7 +263,8 @@ xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
 
 static void
 xkl_config_registry_foreach_in_nodeset(XklConfigRegistry * config,
-				       xmlNodeSetPtr nodes,
+				       GSList ** processed_ids,
+				       gint doc_index, xmlNodeSetPtr nodes,
 				       ConfigItemProcessFunc func,
 				       gpointer data)
 {
@@ -225,8 +273,19 @@ xkl_config_registry_foreach_in_nodeset(XklConfigRegistry * config,
 		xmlNodePtr *pnode = nodes->nodeTab;
 		XklConfigItem *ci = xkl_config_item_new();
 		for (i = nodes->nodeNr; --i >= 0;) {
-			if (xkl_read_config_item(config, *pnode, ci))
-				func(config, ci, data);
+			if (xkl_read_config_item
+			    (config, doc_index, *pnode, ci)) {
+				if (g_slist_find_custom
+				    (*processed_ids, ci->name,
+				     (GCompareFunc) g_ascii_strcasecmp) ==
+				    NULL) {
+					func(config, ci, data);
+					*processed_ids =
+					    g_slist_append(*processed_ids,
+							   g_strdup
+							   (ci->name));
+				}
+			}
 
 			pnode++;
 		}
@@ -242,19 +301,31 @@ xkl_config_registry_foreach_in_xpath(XklConfigRegistry * config,
 				     gpointer data)
 {
 	xmlXPathObjectPtr xpath_obj;
+	gint di;
+	GSList *processed_ids = NULL;
 
 	if (!xkl_config_registry_is_initialized(config))
 		return;
-	xpath_obj = xmlXPathCompiledEval(xpath_comp_expr,
-					 xkl_config_registry_priv
-					 (config, xpath_context));
-	if (xpath_obj != NULL) {
+
+	for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
+		xmlXPathContextPtr xmlctxt =
+		    xkl_config_registry_priv(config, xpath_contexts[di]);
+		if (xmlctxt == NULL)
+			continue;
+
+		xpath_obj = xmlXPathCompiledEval(xpath_comp_expr, xmlctxt);
+		if (xpath_obj == NULL)
+			continue;
+
 		xkl_config_registry_foreach_in_nodeset(config,
+						       &processed_ids, di,
 						       xpath_obj->
-						       nodesetval,
-						       func, data);
+						       nodesetval, func,
+						       data);
 		xmlXPathFreeObject(xpath_obj);
 	}
+	g_slist_foreach(processed_ids, (GFunc) g_free, NULL);
+	g_slist_free(processed_ids);
 }
 
 void
@@ -269,20 +340,34 @@ xkl_config_registry_foreach_in_xpath_with_param(XklConfigRegistry
 {
 	char xpath_expr[1024];
 	xmlXPathObjectPtr xpath_obj;
+	gint di;
+	GSList *processed_ids = NULL;
 
 	if (!xkl_config_registry_is_initialized(config))
 		return;
+
 	g_snprintf(xpath_expr, sizeof xpath_expr, format, value);
-	xpath_obj = xmlXPathEval((unsigned char *) xpath_expr,
-				 xkl_config_registry_priv(config,
-							  xpath_context));
-	if (xpath_obj != NULL) {
+
+	for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
+		xmlXPathContextPtr xmlctxt =
+		    xkl_config_registry_priv(config, xpath_contexts[di]);
+		if (xmlctxt == NULL)
+			continue;
+
+		xpath_obj =
+		    xmlXPathEval((unsigned char *) xpath_expr, xmlctxt);
+		if (xpath_obj == NULL)
+			continue;
+
 		xkl_config_registry_foreach_in_nodeset(config,
+						       &processed_ids, di,
 						       xpath_obj->
-						       nodesetval,
-						       func, data);
+						       nodesetval, func,
+						       data);
 		xmlXPathFreeObject(xpath_obj);
 	}
+	g_slist_foreach(processed_ids, (GFunc) g_free, NULL);
+	g_slist_free(processed_ids);
 }
 
 static gboolean
@@ -296,28 +381,38 @@ xkl_config_registry_find_object(XklConfigRegistry * config,
 	xmlNodeSetPtr nodes;
 	gboolean rv = FALSE;
 	gchar xpath_expr[1024];
+	gint di;
 
 	if (!xkl_config_registry_is_initialized(config))
 		return FALSE;
 
 	g_snprintf(xpath_expr, sizeof xpath_expr, format, arg1,
 		   pitem->name);
-	xpath_obj =
-	    xmlXPathEval((unsigned char *) xpath_expr,
-			 xkl_config_registry_priv(config, xpath_context));
-	if (xpath_obj == NULL)
-		return FALSE;
 
-	nodes = xpath_obj->nodesetval;
-	if (nodes != NULL && nodes->nodeTab != NULL && nodes->nodeNr > 0) {
-		rv = xkl_read_config_item(config, nodes->nodeTab[0],
-					  pitem);
-		if (pnode != NULL) {
-			*pnode = *nodes->nodeTab;
+	for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
+		xmlXPathContextPtr xmlctxt =
+		    xkl_config_registry_priv(config, xpath_contexts[di]);
+		if (xmlctxt == NULL)
+			continue;
+
+		xpath_obj =
+		    xmlXPathEval((unsigned char *) xpath_expr, xmlctxt);
+		if (xpath_obj == NULL)
+			continue;
+
+		nodes = xpath_obj->nodesetval;
+		if (nodes != NULL && nodes->nodeTab != NULL
+		    && nodes->nodeNr > 0) {
+			rv = xkl_read_config_item(config, di,
+						  nodes->nodeTab[0],
+						  pitem);
+			if (pnode != NULL) {
+				*pnode = *nodes->nodeTab;
+			}
 		}
-	}
 
-	xmlXPathFreeObject(xpath_obj);
+		xmlXPathFreeObject(xpath_obj);
+	}
 	return rv;
 }
 
@@ -477,31 +572,33 @@ xkl_xml_sax_end_element_ns(void *ctx,
 
 gboolean
 xkl_config_registry_load_from_file(XklConfigRegistry * config,
-				   const gchar * file_name)
+				   const gchar * file_name, gint docidx)
 {
 	xmlParserCtxtPtr ctxt = xmlNewParserCtxt();
-	xmlSAXHandler *saxh = g_new0(xmlSAXHandler, 1);
+	xmlDocPtr doc;
+
 	xkl_debug(100, "Loading XML registry from file %s\n", file_name);
 
 	/* Filter out all unneeded languages! */
-	xmlSAX2InitDefaultSAXHandler(saxh, TRUE);
-	saxh->startElementNs = xkl_xml_sax_start_element_ns;
-	saxh->endElementNs = xkl_xml_sax_end_element_ns;
-	saxh->characters = xkl_xml_sax_characters;
-	ctxt->sax = saxh;
+	xmlSAX2InitDefaultSAXHandler(ctxt->sax, TRUE);
+	ctxt->sax->startElementNs = xkl_xml_sax_start_element_ns;
+	ctxt->sax->endElementNs = xkl_xml_sax_end_element_ns;
+	ctxt->sax->characters = xkl_xml_sax_characters;
 
-	xkl_config_registry_priv(config, doc) =
+	doc = xkl_config_registry_priv(config, docs[docidx]) =
 	    xmlCtxtReadFile(ctxt, file_name, NULL, XML_PARSE_NOBLANKS);
 	xmlFreeParserCtxt(ctxt);
 
-	if (xkl_config_registry_priv(config, doc) == NULL) {
-		xkl_config_registry_priv(config, xpath_context) = NULL;
+	if (doc == NULL) {
+		xkl_config_registry_priv(config, xpath_contexts[docidx]) =
+		    NULL;
 		xkl_last_error_message =
-		    "Could not parse XKB configuration registry";
+		    "Could not parse primary XKB configuration registry";
 		return FALSE;
 	}
-	xkl_config_registry_priv(config, xpath_context) =
-	    xmlXPathNewContext(xkl_config_registry_priv(config, doc));
+
+	xkl_config_registry_priv(config, xpath_contexts[docidx]) =
+	    xmlXPathNewContext(doc);
 
 	return TRUE;
 }
@@ -509,7 +606,8 @@ xkl_config_registry_load_from_file(XklConfigRegistry * config,
 gboolean
 xkl_config_registry_load_helper(XklConfigRegistry * config,
 				const char default_ruleset[],
-				const char base_dir[])
+				const char base_dir[],
+				gboolean if_extras_needed)
 {
 	struct stat stat_buf;
 	gchar file_name[MAXPATHLEN] = "";
@@ -527,18 +625,43 @@ xkl_config_registry_load_helper(XklConfigRegistry * config,
 		return FALSE;
 	}
 
-	return xkl_config_registry_load_from_file(config, file_name);
+	if (!xkl_config_registry_load_from_file(config, file_name, 0))
+		return FALSE;
+
+	if (!if_extras_needed)
+		return TRUE;
+
+	g_snprintf(file_name, sizeof file_name, "%s/%s.extras.xml",
+		   base_dir, rf);
+
+	/* no extras - ok, no problem */
+	if (stat(file_name, &stat_buf) != 0)
+		return TRUE;
+
+	return xkl_config_registry_load_from_file(config, file_name, 1);
 }
 
 void
 xkl_config_registry_free(XklConfigRegistry * config)
 {
 	if (xkl_config_registry_is_initialized(config)) {
-		xmlXPathFreeContext(xkl_config_registry_priv
-				    (config, xpath_context));
-		xmlFreeDoc(xkl_config_registry_priv(config, doc));
-		xkl_config_registry_priv(config, xpath_context) = NULL;
-		xkl_config_registry_priv(config, doc) = NULL;
+		gint di;
+		for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
+			xmlXPathContextPtr xmlctxt =
+			    xkl_config_registry_priv(config,
+						     xpath_contexts[di]);
+			if (xmlctxt == NULL)
+				continue;
+
+			xmlXPathFreeContext(xmlctxt);
+			xmlFreeDoc(xkl_config_registry_priv
+				   (config, docs[di]));
+			xkl_config_registry_priv(config,
+						 xpath_contexts[di]) =
+			    NULL;
+			xkl_config_registry_priv(config, docs[di]) = NULL;
+		}
+
 	}
 }
 
@@ -582,40 +705,61 @@ xkl_config_registry_foreach_option_group(XklConfigRegistry *
 					 func, gpointer data)
 {
 	xmlXPathObjectPtr xpath_obj;
-	gint i;
+	gint di, j;
+	GSList *processed_ids = NULL;
 
 	if (!xkl_config_registry_is_initialized(config))
 		return;
-	xpath_obj =
-	    xmlXPathCompiledEval(option_groups_xpath,
-				 xkl_config_registry_priv(config,
-							  xpath_context));
-	if (xpath_obj != NULL) {
-		xmlNodeSetPtr nodes = xpath_obj->nodesetval;
-		xmlNodePtr *pnode = nodes->nodeTab;
-		XklConfigItem *ci = xkl_config_item_new();
-		for (i = nodes->nodeNr; --i >= 0;) {
 
-			if (xkl_read_config_item(config, *pnode, ci)) {
-				gboolean allow_multisel = TRUE;
-				xmlChar *sallow_multisel =
-				    xmlGetProp(*pnode,
-					       (unsigned char *)
-					       XCI_PROP_ALLOW_MULTIPLE_SELECTION);
-				if (sallow_multisel != NULL) {
-					allow_multisel =
-					    !g_ascii_strcasecmp
-					    ("true", (char *)
-					     sallow_multisel);
-					xmlFree(sallow_multisel);
-					g_object_set_data(G_OBJECT
-							  (ci),
-							  XCI_PROP_ALLOW_MULTIPLE_SELECTION,
-							  GINT_TO_POINTER
-							  (allow_multisel));
+	for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
+		xmlNodeSetPtr nodes;
+		xmlNodePtr *pnode;
+		XklConfigItem *ci;
+
+		xmlXPathContextPtr xmlctxt =
+		    xkl_config_registry_priv(config, xpath_contexts[di]);
+		if (xmlctxt == NULL)
+			continue;
+
+		xpath_obj =
+		    xmlXPathCompiledEval(option_groups_xpath, xmlctxt);
+		if (xpath_obj == NULL)
+			continue;
+
+		nodes = xpath_obj->nodesetval;
+		pnode = nodes->nodeTab;
+		ci = xkl_config_item_new();
+		for (j = nodes->nodeNr; --j >= 0;) {
+
+			if (xkl_read_config_item(config, di, *pnode, ci)) {
+				if (g_slist_find_custom
+				    (processed_ids, ci->name,
+				     (GCompareFunc) g_ascii_strcasecmp) ==
+				    NULL) {
+					gboolean allow_multisel = TRUE;
+					xmlChar *sallow_multisel =
+					    xmlGetProp(*pnode,
+						       (unsigned char *)
+						       XCI_PROP_ALLOW_MULTIPLE_SELECTION);
+					if (sallow_multisel != NULL) {
+						allow_multisel =
+						    !g_ascii_strcasecmp
+						    ("true", (char *)
+						     sallow_multisel);
+						xmlFree(sallow_multisel);
+						g_object_set_data(G_OBJECT
+								  (ci),
+								  XCI_PROP_ALLOW_MULTIPLE_SELECTION,
+								  GINT_TO_POINTER
+								  (allow_multisel));
+					}
+
+					func(config, ci, data);
+					processed_ids =
+					    g_slist_append(processed_ids,
+							   g_strdup
+							   (ci->name));
 				}
-
-				func(config, ci, data);
 			}
 
 			pnode++;
@@ -623,6 +767,8 @@ xkl_config_registry_foreach_option_group(XklConfigRegistry *
 		g_object_unref(G_OBJECT(ci));
 		xmlXPathFreeObject(xpath_obj);
 	}
+	g_slist_foreach(processed_ids, (GFunc) g_free, NULL);
+	g_slist_free(processed_ids);
 }
 
 void
@@ -639,9 +785,197 @@ xkl_config_registry_foreach_option(XklConfigRegistry * config,
 							func, data);
 }
 
+static gboolean
+search_all(const gchar * haystack, gchar ** needles)
+{
+	/* match anything */
+	if (!needles || !*needles)
+		return TRUE;
+
+	gchar *uchs = g_utf8_strup(haystack, -1);
+	do {
+		if (g_strstr_len(uchs, -1, *needles) == NULL) {
+			g_free(uchs);
+			return FALSE;
+		}
+		needles++;
+	} while (*needles);
+
+	g_free(uchs);
+	return TRUE;
+}
+
+static gboolean
+if_country_matches_pattern(const XklConfigItem * item,
+			   gchar ** patterns, const gboolean check_name)
+{
+	const gchar *country_desc;
+	if (check_name) {
+		gchar *upper_name = g_ascii_strup(item->name, -1);
+		country_desc = xkl_get_country_name(upper_name);
+		g_free(upper_name);
+		xkl_debug(200, "Checking layout country: [%s]\n",
+			  country_desc);
+		if ((country_desc != NULL)
+		    && search_all(country_desc, patterns))
+			return TRUE;
+	}
+
+	gchar **countries = g_object_get_data(G_OBJECT(item),
+					      XCI_PROP_COUNTRY_LIST);
+	for (; countries && *countries; countries++) {
+		country_desc = xkl_get_country_name(*countries);
+		xkl_debug(200, "Checking country: [%s][%s]\n",
+			  *countries, country_desc);
+		if ((country_desc != NULL)
+		    && search_all(country_desc, patterns)) {
+			return TRUE;
+		}
+	}
+	return FALSE;
+}
+
+static gboolean
+if_language_matches_pattern(const XklConfigItem * item,
+			    gchar ** patterns, const gboolean check_name)
+{
+	const gchar *language_desc;
+	if (check_name) {
+		language_desc = xkl_get_language_name(item->name);
+		xkl_debug(200, "Checking layout language: [%s]\n",
+			  language_desc);
+		if ((language_desc != NULL)
+		    && search_all(language_desc, patterns))
+			return TRUE;
+	}
+	gchar **languages = g_object_get_data(G_OBJECT(item),
+					      XCI_PROP_LANGUAGE_LIST);
+	for (; languages && *languages; languages++) {
+		language_desc = xkl_get_language_name(*languages);
+		xkl_debug(200, "Checking language: [%s][%s]\n",
+			  *languages, language_desc);
+		if ((language_desc != NULL)
+		    && search_all(language_desc, patterns)) {
+			return TRUE;
+		}
+	}
+	return FALSE;
+}
+
+static void
+xkl_config_registry_search_by_pattern_in_variant(XklConfigRegistry *
+						 config,
+						 const XklConfigItem *
+						 item,
+						 SearchParamType *
+						 search_param)
+{
+	gboolean variant_matched = FALSE;
+	gchar *full_desc = g_strdup_printf("%s - %s",
+					   search_param->
+					   layout_item->description,
+					   item->description);
+
+	xkl_debug(200, "Variant to check: [%s][%s]\n", item->name,
+		  item->description);
+
+	if (search_all(full_desc, search_param->patterns))
+		variant_matched = TRUE;
+
+	g_free(full_desc);
+
+	if (!variant_matched) {
+		gchar **countries = g_object_get_data(G_OBJECT(item),
+						      XCI_PROP_COUNTRY_LIST);
+		if (countries && g_strv_length(countries) > 0) {
+			if (if_country_matches_pattern
+			    (item, search_param->patterns, FALSE))
+				variant_matched = TRUE;
+		} else {
+			if (search_param->country_matched)
+				variant_matched = TRUE;
+		}
+	}
+
+	if (!variant_matched) {
+		gchar **languages = g_object_get_data(G_OBJECT(item),
+						      XCI_PROP_LANGUAGE_LIST);
+		if (languages && g_strv_length(languages) > 0) {
+			if (if_language_matches_pattern
+			    (item, search_param->patterns, FALSE))
+				variant_matched = TRUE;
+		} else {
+			if (search_param->language_matched)
+				variant_matched = TRUE;
+		}
+	}
+
+	if (variant_matched)
+		(search_param->func) (config, search_param->layout_item,
+				      item, search_param->data);
+}
+
+static void
+xkl_config_registry_search_by_pattern_in_layout(XklConfigRegistry * config,
+						const XklConfigItem * item,
+						SearchParamType *
+						search_param)
+{
+	gchar *upper_name = g_ascii_strup(item->name, -1);
+
+	xkl_debug(200, "Layout to check: [%s][%s]\n", item->name,
+		  item->description);
+
+	search_param->country_matched =
+	    search_param->language_matched = FALSE;
+
+	if (if_country_matches_pattern(item, search_param->patterns, TRUE))
+		search_param->country_matched = TRUE;
+	else if (if_language_matches_pattern
+		 (item, search_param->patterns, TRUE))
+		search_param->language_matched = TRUE;
+	else if (search_all(item->description, search_param->patterns))
+		search_param->language_matched = TRUE;
+
+	if (search_param->country_matched
+	    || search_param->language_matched)
+		(search_param->func) (config, item, NULL,
+				      search_param->data);
+
+	search_param->layout_item = item;
+
+	xkl_config_registry_foreach_layout_variant(config, item->name,
+						   (ConfigItemProcessFunc)
+						   xkl_config_registry_search_by_pattern_in_variant,
+						   search_param);
+
+	g_free(upper_name);
+}
+
+void
+xkl_config_registry_search_by_pattern(XklConfigRegistry
+				      * config,
+				      const gchar *
+				      pattern,
+				      TwoConfigItemsProcessFunc
+				      func, gpointer data)
+{
+	xkl_debug(200, "Searching by pattern: [%s]\n", pattern);
+	gchar *upattern = pattern ? g_utf8_strup(pattern, -1) : NULL;
+	gchar **patterns = pattern ? g_strsplit(upattern, " ", -1) : NULL;
+	SearchParamType search_param = {
+		patterns, func, data
+	};
+	xkl_config_registry_foreach_layout(config, (ConfigItemProcessFunc)
+					   xkl_config_registry_search_by_pattern_in_layout,
+					   &search_param);
+	g_strfreev(patterns);
+	g_free(upattern);
+}
+
 gboolean
-xkl_config_registry_find_model(XklConfigRegistry * config,
-			       XklConfigItem * pitem /* in/out */ )
+xkl_config_registry_find_model(XklConfigRegistry *
+			       config, XklConfigItem * pitem /* in/out */ )
 {
 	return xkl_config_registry_find_object(config,
 					       XKBCR_MODEL_PATH
@@ -650,7 +984,8 @@ xkl_config_registry_find_model(XklConfigRegistry * config,
 }
 
 gboolean
-xkl_config_registry_find_layout(XklConfigRegistry * config,
+xkl_config_registry_find_layout(XklConfigRegistry *
+				config,
 				XklConfigItem * pitem /* in/out */ )
 {
 	return xkl_config_registry_find_object(config,
@@ -660,8 +995,10 @@ xkl_config_registry_find_layout(XklConfigRegistry * config,
 }
 
 gboolean
-xkl_config_registry_find_variant(XklConfigRegistry * config,
-				 const char *layout_name,
+xkl_config_registry_find_variant(XklConfigRegistry *
+				 config,
+				 const char
+				 *layout_name,
 				 XklConfigItem * pitem /* in/out */ )
 {
 	return xkl_config_registry_find_object(config,
@@ -671,21 +1008,16 @@ xkl_config_registry_find_variant(XklConfigRegistry * config,
 }
 
 gboolean
-xkl_config_registry_find_option_group(XklConfigRegistry *
-				      config,
-				      XklConfigItem * pitem /* in/out */ )
+xkl_config_registry_find_option_group(XklConfigRegistry * config, XklConfigItem * pitem	/* in/out */
+    )
 {
 	xmlNodePtr node = NULL;
 	gboolean rv = xkl_config_registry_find_object(config,
 						      XKBCR_GROUP_PATH
 						      "[configItem/name = '%s%s']",
-						      "",
-						      pitem,
-						      &node);
-
+						      "", pitem, &node);
 	if (rv) {
-		xmlChar *val = xmlGetProp(node,
-					  (unsigned char *)
+		xmlChar *val = xmlGetProp(node, (unsigned char *)
 					  XCI_PROP_ALLOW_MULTIPLE_SELECTION);
 		if (val != NULL) {
 			gboolean allow_multisel =
@@ -701,8 +1033,10 @@ xkl_config_registry_find_option_group(XklConfigRegistry *
 }
 
 gboolean
-xkl_config_registry_find_option(XklConfigRegistry * config,
-				const char *option_group_name,
+xkl_config_registry_find_option(XklConfigRegistry *
+				config,
+				const char
+				*option_group_name,
 				XklConfigItem * pitem /* in/out */ )
 {
 	return xkl_config_registry_find_object(config,
@@ -724,14 +1058,16 @@ xkl_config_rec_activate(const XklConfigRec * data, XklEngine * engine)
 }
 
 gboolean
-xkl_config_registry_load(XklConfigRegistry * config)
+xkl_config_registry_load(XklConfigRegistry * config,
+			 gboolean if_extras_needed)
 {
 	XklEngine *engine;
-
 	xkl_config_registry_free(config);
 	engine = xkl_config_registry_get_engine(config);
 	xkl_engine_ensure_vtable_inited(engine);
-	return xkl_engine_vcall(engine, load_config_registry) (config);
+	return xkl_engine_vcall(engine,
+				load_config_registry) (config,
+						       if_extras_needed);
 }
 
 gboolean
@@ -751,10 +1087,8 @@ xkl_config_rec_write_to_file(XklEngine * engine,
 		return FALSE;
 	}
 	xkl_engine_ensure_vtable_inited(engine);
-	return xkl_engine_vcall(engine,
-				write_config_rec_to_file) (engine,
-							   file_name,
-							   data, binary);
+	return xkl_engine_vcall(engine, write_config_rec_to_file)
+	    (engine, file_name, data, binary);
 }
 
 void
@@ -762,7 +1096,6 @@ xkl_config_rec_dump(FILE * file, XklConfigRec * data)
 {
 	int j;
 	fprintf(file, "  model: [%s]\n", data->model);
-
 	fprintf(file, "  layouts:\n");
 #define OUTPUT_ARRZ(arrz) \
   { \
@@ -775,7 +1108,6 @@ xkl_config_rec_dump(FILE * file, XklConfigRec * data)
 	OUTPUT_ARRZ(layouts);
 	OUTPUT_ARRZ(variants);
 	OUTPUT_ARRZ(options);
-
 }
 
 G_DEFINE_TYPE(XklConfigRegistry, xkl_config_registry, G_TYPE_OBJECT)
@@ -783,14 +1115,12 @@ static GObject *
 xkl_config_registry_constructor(GType type,
 				guint
 				n_construct_properties,
-				GObjectConstructParam
-				* construct_properties)
+				GObjectConstructParam *
+				construct_properties)
 {
 	GObject *obj;
 	XklConfigRegistry *config;
-	XklEngine *engine;
-
-	{
+	XklEngine *engine; {
 		/* Invoke parent constructor. */
 		XklConfigRegistryClass *klass;
 		klass =
@@ -803,14 +1133,11 @@ xkl_config_registry_constructor(GType type,
 	}
 
 	config = XKL_CONFIG_REGISTRY(obj);
-
 	engine = XKL_ENGINE(g_value_peek_pointer
 			    (construct_properties[0].value));
 	xkl_config_registry_get_engine(config) = engine;
-
 	xkl_engine_ensure_vtable_inited(engine);
 	xkl_engine_vcall(engine, init_config_registry) (config);
-
 	return obj;
 }
 
@@ -833,7 +1160,6 @@ xkl_config_registry_get_property(GObject * object,
 				 GValue * value, GParamSpec * pspec)
 {
 	XklConfigRegistry *config = XKL_CONFIG_REGISTRY(object);
-
 	switch (property_id) {
 	case PROP_ENGINE:
 		g_value_set_pointer(value,
@@ -848,11 +1174,8 @@ static void
 xkl_config_registry_finalize(GObject * obj)
 {
 	XklConfigRegistry *config = (XklConfigRegistry *) obj;
-
 	xkl_config_registry_free(config);
-
 	g_free(config->priv);
-
 	G_OBJECT_CLASS(parent_class)->finalize(obj);
 }
 
@@ -864,6 +1187,7 @@ xkl_config_registry_finalize(GObject * obj)
 extern void
 xkl_config_registry_class_term(XklConfigRegistryClass * klass)
 {
+	gint i;
 	if (models_xpath != NULL) {
 		xmlXPathFreeCompExpr(models_xpath);
 		models_xpath = NULL;
@@ -876,6 +1200,24 @@ xkl_config_registry_class_term(XklConfigRegistryClass * klass)
 		xmlXPathFreeCompExpr(option_groups_xpath);
 		option_groups_xpath = NULL;
 	}
+	if (xml_encode_regexen != NULL) {
+		for (i =
+		     sizeof(xml_encode_regexen_str) /
+		     sizeof(xml_encode_regexen_str[0]); --i >= 0;) {
+			g_regex_unref(xml_encode_regexen[i]);
+		}
+		g_free(xml_encode_regexen);
+		xml_encode_regexen = NULL;
+	}
+	if (xml_decode_regexen != NULL) {
+		for (i =
+		     sizeof(xml_decode_regexen_str) /
+		     sizeof(xml_decode_regexen_str[0]); --i >= 0;) {
+			g_regex_unref(xml_decode_regexen[i]);
+		}
+		g_free(xml_decode_regexen);
+		xml_decode_regexen = NULL;
+	}
 }
 
 static void
@@ -883,28 +1225,22 @@ xkl_config_registry_class_init(XklConfigRegistryClass * klass)
 {
 	GObjectClass *object_class;
 	GParamSpec *engine_param_spec;
-
+	gint i;
 	object_class = (GObjectClass *) klass;
 	parent_class = g_type_class_peek_parent(object_class);
 	object_class->constructor = xkl_config_registry_constructor;
 	object_class->finalize = xkl_config_registry_finalize;
 	object_class->set_property = xkl_config_registry_set_property;
 	object_class->get_property = xkl_config_registry_get_property;
-
-	bind_textdomain_codeset (XKB_DOMAIN, "UTF-8");
-
-	engine_param_spec = g_param_spec_object("engine",
-						"Engine",
-						"XklEngine",
-						XKL_TYPE_ENGINE,
-						G_PARAM_CONSTRUCT_ONLY
-						| G_PARAM_READWRITE);
-
-	g_object_class_install_property(object_class,
-					PROP_ENGINE, engine_param_spec);
-
+	bind_textdomain_codeset(XKB_DOMAIN, "UTF-8");
+	engine_param_spec =
+	    g_param_spec_object("engine", "Engine", "XklEngine",
+				XKL_TYPE_ENGINE,
+				G_PARAM_CONSTRUCT_ONLY |
+				G_PARAM_READWRITE);
+	g_object_class_install_property(object_class, PROP_ENGINE,
+					engine_param_spec);
 	/* static stuff initialized */
-
 	xmlXPathInit();
 	models_xpath = xmlXPathCompile((unsigned char *)
 				       XKBCR_MODEL_PATH);
@@ -912,4 +1248,20 @@ xkl_config_registry_class_init(XklConfigRegistryClass * klass)
 					XKBCR_LAYOUT_PATH);
 	option_groups_xpath = xmlXPathCompile((unsigned char *)
 					      XKBCR_GROUP_PATH);
+	xml_encode_regexen =
+	    g_new0(GRegex *,
+		   sizeof(xml_encode_regexen_str) /
+		   sizeof(xml_encode_regexen_str[0]));
+	xml_decode_regexen =
+	    g_new0(GRegex *,
+		   sizeof(xml_decode_regexen_str) /
+		   sizeof(xml_decode_regexen_str[0]));
+	for (i =
+	     sizeof(xml_encode_regexen_str) /
+	     sizeof(xml_encode_regexen_str[0]); --i >= 0;) {
+		xml_encode_regexen[i] =
+		    g_regex_new(xml_encode_regexen_str[i], 0, 0, NULL);
+		xml_decode_regexen[i] =
+		    g_regex_new(xml_decode_regexen_str[i], 0, 0, NULL);
+	}
 }
