@@ -50,8 +50,7 @@ xkl_engine_filter_events(XklEngine * engine, XEvent * xev)
 			break;
 		case CreateNotify:
 			xkl_engine_process_create_window_evt(engine,
-							     &xev->
-							     xcreatewindow);
+							     &xev->xcreatewindow);
 			break;
 		case DestroyNotify:
 			xkl_debug(150,
@@ -101,12 +100,14 @@ xkl_engine_process_focus_in_evt(XklEngine * engine,
 				XFocusChangeEvent * fev)
 {
 	Window win;
+	Window prev_toplevel_win =
+	    xkl_engine_priv(engine, curr_toplvl_win);
 	Window toplevel_win;
 	XklState selected_window_state;
 
 	if (!
-	    (xkl_engine_priv(engine, listener_type) &
-	     XKLL_MANAGE_WINDOW_STATES))
+	    (xkl_engine_is_listening_for
+	     (engine, XKLL_MANAGE_WINDOW_STATES)))
 		return;
 
 	win = fev->window;
@@ -136,38 +137,51 @@ xkl_engine_process_focus_in_evt(XklEngine * engine,
 
 	if (xkl_engine_get_toplevel_window_state
 	    (engine, toplevel_win, &selected_window_state)) {
-		if (xkl_engine_priv(engine, curr_toplvl_win) !=
-		    toplevel_win) {
-			gboolean old_win_transparent, new_win_transparent;
-			XklState tmp_state;
-
-			old_win_transparent =
-			    xkl_engine_is_toplevel_window_transparent
-			    (engine,
-			     xkl_engine_priv(engine, curr_toplvl_win));
-			if (old_win_transparent)
-				xkl_debug(150,
-					  "Leaving transparent window\n");
+		if (prev_toplevel_win != toplevel_win) {
+			gboolean new_win_transparent;
+			Window parent = (Window) NULL, root =
+			    (Window) NULL, *children = NULL;
+			guint nchildren = 0;
 
 			/*
-			 * Reload the current state from the current window. 
-			 * Do not do it for transparent window - we keep the state from 
-			 * the _previous_ window.
+			 * If previous focused window exists - handle transparency and state
+			 * (optional)
 			 */
-			if (!old_win_transparent &&
-			    xkl_engine_get_toplevel_window_state(engine,
-								 xkl_engine_priv
-								 (engine,
-								  curr_toplvl_win),
-								 &tmp_state))
-			{
-				xkl_engine_update_current_state(engine,
-								tmp_state.
-								group,
-								tmp_state.
-								indicators,
-								"Loading current (previous) state from the current (previous) window");
-			}
+			if (xkl_engine_query_tree
+			    (engine, prev_toplevel_win, &root, &parent,
+			     &children, &nchildren) == Success) {
+				XklState tmp_state;
+				gboolean old_win_transparent =
+				    xkl_engine_is_toplevel_window_transparent
+				    (engine, prev_toplevel_win);
+
+				if (children != NULL)
+					XFree(children);
+
+				if (old_win_transparent)
+					xkl_debug(150,
+						  "Leaving transparent window\n");
+				/*
+				 * Reload the current state from the current window. 
+				 * Do not do it for transparent window - we keep the state from 
+				 * the _previous_ window.
+				 */
+				if (!old_win_transparent
+				    &&
+				    xkl_engine_get_toplevel_window_state
+				    (engine, prev_toplevel_win,
+				     &tmp_state)) {
+					xkl_engine_update_current_state
+					    (engine, tmp_state.group,
+					     tmp_state.indicators,
+					     "Loading current (previous) state from the current (previous) window");
+				}
+			} else
+				xkl_debug(150,
+					  "Current (previous) window "
+					  WINID_FORMAT
+					  " does not exist any more, so transparency/state are not analyzed\n",
+					  prev_toplevel_win);
 
 			xkl_engine_priv(engine, curr_toplvl_win) =
 			    toplevel_win;
@@ -219,10 +233,8 @@ xkl_engine_process_focus_in_evt(XklEngine * engine,
 							  "Restoring the group from %d to %d after gaining focus\n",
 							  xkl_engine_priv
 							  (engine,
-							   curr_state).
-							  group,
-							  selected_window_state.
-							  group);
+							   curr_state).group,
+							  selected_window_state.group);
 						/*
 						 *  For fast mouse movements - the state is probably not updated yet
 						 *  (because of the group change notification being late).
@@ -230,21 +242,18 @@ xkl_engine_process_focus_in_evt(XklEngine * engine,
 						 */
 						xkl_engine_update_current_state
 						    (engine,
-						     selected_window_state.
-						     group,
-						     selected_window_state.
-						     indicators,
+						     selected_window_state.group,
+						     selected_window_state.indicators,
 						     "Enforcing fast update of the current state");
 						xkl_engine_lock_group
 						    (engine,
-						     selected_window_state.
-						     group);
+						     selected_window_state.group);
+						xkl_engine_priv(engine, skip_one_save) = TRUE;
 					} else {
 						xkl_debug(150,
 							  "Both old and new focused window "
 							  "have group %d so no point restoring it\n",
-							  selected_window_state.
-							  group);
+							  selected_window_state.group);
 						xkl_engine_one_switch_to_secondary_group_performed
 						    (engine);
 					}
@@ -258,10 +267,8 @@ xkl_engine_process_focus_in_evt(XklEngine * engine,
 					xkl_debug(150,
 						  "Restoring the indicators from %X to %X after gaining focus\n",
 						  xkl_engine_priv(engine,
-								  curr_state).
-						  indicators,
-						  selected_window_state.
-						  indicators);
+								  curr_state).indicators,
+						  selected_window_state.indicators);
 					xkl_engine_ensure_vtable_inited
 					    (engine);
 					xkl_engine_vcall(engine,
@@ -272,14 +279,12 @@ xkl_engine_process_focus_in_evt(XklEngine * engine,
 					xkl_debug(150,
 						  "Not restoring the indicators %X after gaining focus: indicator handling is not enabled\n",
 						  xkl_engine_priv(engine,
-								  curr_state).
-						  indicators);
+								  curr_state).indicators);
 			} else
 				xkl_debug(150,
 					  "Not restoring the group %d after gaining focus: global layout (xor transparent window)\n",
 					  xkl_engine_priv(engine,
-							  curr_state).
-					  group);
+							  curr_state).group);
 		} else
 			xkl_debug(150,
 				  "Same app window - just do nothing\n");
@@ -321,8 +326,8 @@ xkl_engine_process_focus_out_evt(XklEngine * engine,
 				 XFocusChangeEvent * fev)
 {
 	if (!
-	    (xkl_engine_priv(engine, listener_type) &
-	     XKLL_MANAGE_WINDOW_STATES))
+	    (xkl_engine_is_listening_for
+	     (engine, XKLL_MANAGE_WINDOW_STATES)))
 		return;
 
 	if (fev->mode != NotifyNormal) {
@@ -365,7 +370,6 @@ xkl_engine_process_focus_out_evt(XklEngine * engine,
 void
 xkl_engine_process_property_evt(XklEngine * engine, XPropertyEvent * pev)
 {
-	guint listener_type = xkl_engine_priv(engine, listener_type);
 	if (400 <= xkl_debug_level) {
 		char *atom_name =
 		    XGetAtomName(xkl_engine_get_display(engine),
@@ -384,7 +388,8 @@ xkl_engine_process_property_evt(XklEngine * engine, XPropertyEvent * pev)
 	}
 
 	if (pev->atom == xkl_engine_priv(engine, atoms)[WM_STATE]) {
-		if (listener_type & XKLL_MANAGE_WINDOW_STATES) {
+		if (xkl_engine_is_listening_for
+		    (engine, XKLL_MANAGE_WINDOW_STATES)) {
 			gboolean has_xkl_state =
 			    xkl_engine_get_state(engine, pev->window,
 						 NULL);
@@ -405,21 +410,22 @@ xkl_engine_process_property_evt(XklEngine * engine, XPropertyEvent * pev)
 					  "Something (%d) happened to WM_STATE of window 0x%x\n",
 					  pev->state, pev->window);
 				xkl_engine_select_input_merging(engine,
-								pev->
-								window,
+								pev->window,
 								PropertyChangeMask);
 				if (has_xkl_state) {
 					xkl_engine_delete_state(engine,
-								pev->
-								window);
+								pev->window);
 				}
 			}
 		}		/* XKLL_MANAGE_WINDOW_STATES */
 	} else if (pev->atom == xkl_engine_priv(engine, base_config_atom)
 		   && pev->window == xkl_engine_priv(engine, root_window)) {
-		if (listener_type &
-		    (XKLL_MANAGE_WINDOW_STATES |
-		     XKLL_TRACK_KEYBOARD_STATE)) {
+		if (xkl_engine_is_listening_for
+		    (engine,
+		     XKLL_MANAGE_WINDOW_STATES) |
+		    xkl_engine_is_listening_for(engine,
+						XKLL_TRACK_KEYBOARD_STATE))
+		{
 			if (pev->state == PropertyNewValue) {
 				/* If root window got new *_NAMES_PROP_ATOM -
 				   it most probably means new keyboard config is loaded by somebody */
@@ -438,9 +444,8 @@ void
 xkl_engine_process_create_window_evt(XklEngine * engine,
 				     XCreateWindowEvent * cev)
 {
-	if (!
-	    (xkl_engine_priv(engine, listener_type) &
-	     XKLL_MANAGE_WINDOW_STATES))
+	if (!xkl_engine_is_listening_for
+	    (engine, XKLL_MANAGE_WINDOW_STATES))
 		return;
 
 	xkl_debug(200,
@@ -477,20 +482,25 @@ xkl_engine_process_create_window_evt(XklEngine * engine,
 /*
  * Just error handler - sometimes we get BadWindow error for already gone 
  * windows, so we'll just ignore
+ * This handler can be called in the middle of the engine initialization -
+ * so it is not fair to assume that the engine is available
  */
-void
+int
 xkl_process_error(Display * dpy, XErrorEvent * evt)
 {
 	char buf[128] = "";
 	XklEngine *engine = xkl_get_the_engine();
 
-	xkl_engine_priv(engine, last_error_code) = evt->error_code;
-	switch (xkl_engine_priv(engine, last_error_code)) {
-	case BadWindow:
+	if (engine != NULL)
+		xkl_engine_priv(engine, last_error_code) = evt->error_code;
+
+	switch (evt->error_code) {
 	case BadAccess:
+	case BadDrawable:
+	case BadWindow:
 		{
-			XGetErrorText(xkl_engine_get_display(engine),
-				      evt->error_code, buf, sizeof(buf));
+			XGetErrorText(evt->display, evt->error_code, buf,
+				      sizeof(buf));
 			/* in most cases this means we are late:) */
 			xkl_debug(200,
 				  "ERROR: %p, " WINID_FORMAT ", %d [%s], "
@@ -502,6 +512,22 @@ xkl_process_error(Display * dpy, XErrorEvent * evt)
 			break;
 		}
 	default:
+		if (engine != NULL
+		    && xkl_engine_priv(engine, process_x_error)) {
+			if (xkl_engine_priv(engine, process_x_error)
+			    (engine, evt)) {
+				xkl_debug(200,
+					  "X ERROR processed by the engine: %p, "
+					  WINID_FORMAT ", %d [%s], "
+					  "X11 request: %d, minor code: %d\n",
+					  dpy,
+					  (unsigned long) evt->resourceid,
+					  (int) evt->error_code, buf,
+					  (int) evt->request_code,
+					  (int) evt->minor_code);
+				break;
+			}
+		}
 		xkl_debug(200,
 			  "Unexpected by libxklavier X ERROR: %p, "
 			  WINID_FORMAT ", %d [%s], "
@@ -509,10 +535,15 @@ xkl_process_error(Display * dpy, XErrorEvent * evt)
 			  (unsigned long) evt->resourceid,
 			  (int) evt->error_code, buf,
 			  (int) evt->request_code, (int) evt->minor_code);
-		if (!xkl_engine_priv(engine, criticalSection))
-			(*xkl_engine_priv(engine, default_error_handler))
-			    (dpy, evt);
+		if (engine != NULL)
+			if (!xkl_engine_priv(engine, criticalSection))
+				(*xkl_engine_priv
+				 (engine, default_error_handler))
+				    (dpy, evt);
 	}
+
+	/* X ignores this return value anyway */
+	return 0;
 }
 
 /*
@@ -530,6 +561,12 @@ xkl_engine_process_state_modification(XklEngine * engine,
 	gboolean have_old_state = TRUE;
 	gboolean set_group = change_type == GROUP_CHANGED;
 
+	if (xkl_engine_priv(engine, skip_one_save)) {
+		xkl_debug(160, "Skipping one callback");
+		xkl_engine_priv(engine, skip_one_save) = FALSE;
+		return;
+	}
+
 	XGetInputFocus(xkl_engine_get_display(engine), &focused, &revert);
 
 	if ((focused == None) || (focused == PointerRoot)) {
@@ -543,15 +580,14 @@ xkl_engine_process_state_modification(XklEngine * engine,
 	 */
 	if (!xkl_engine_find_toplevel_window
 	    (engine, focused, &focused_toplevel)
-	    && xkl_engine_priv(engine,
-			       listener_type) & XKLL_MANAGE_WINDOW_STATES)
+	    && xkl_engine_is_listening_for(engine,
+					   XKLL_MANAGE_WINDOW_STATES))
 		focused_toplevel = xkl_engine_priv(engine, curr_toplvl_win);	/* what else can I do */
 
 	xkl_debug(150, "Focused window: " WINID_FORMAT ", '%s'\n",
 		  focused_toplevel,
 		  xkl_get_debug_window_title(engine, focused_toplevel));
-	if (xkl_engine_priv(engine, listener_type) &
-	    XKLL_MANAGE_WINDOW_STATES) {
+	if (xkl_engine_is_listening_for(engine, XKLL_MANAGE_WINDOW_STATES)) {
 		xkl_debug(150, "CurClient: " WINID_FORMAT ", '%s'\n",
 			  xkl_engine_priv(engine, curr_toplvl_win),
 			  xkl_get_debug_window_title(engine,
@@ -569,8 +605,8 @@ xkl_engine_process_state_modification(XklEngine * engine,
 				xkl_engine_update_current_state(engine,
 								grp, inds,
 								"Updating the state from new focused window");
-				if (xkl_engine_priv(engine, listener_type)
-				    & XKLL_MANAGE_WINDOW_STATES)
+				if (xkl_engine_is_listening_for
+				    (engine, XKLL_MANAGE_WINDOW_STATES))
 					xkl_engine_add_toplevel_window
 					    (engine, focused_toplevel,
 					     (Window) NULL, FALSE,
@@ -624,8 +660,7 @@ xkl_engine_process_state_modification(XklEngine * engine,
 		xkl_engine_try_call_state_func(engine, change_type,
 					       &old_state);
 
-	if (xkl_engine_priv(engine, listener_type) &
-	    XKLL_MANAGE_WINDOW_STATES)
+	if (xkl_engine_is_listening_for(engine, XKLL_MANAGE_WINDOW_STATES))
 		xkl_engine_save_toplevel_window_state(engine,
 						      xkl_engine_priv
 						      (engine,
